﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System.Collections.Generic;
using System.Linq;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.PermutationEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.TravelingSalesman;

namespace HeuristicLab.Analysis.AlgorithmBehavior {

  /// <summary>
  /// An operator that analyzes the quality of schemas.
  /// </summary>
  [Item("DistinctSolutionsAnalyzer", "An operator that analyzes broken inheritance of building blocks.")]
  [StorableClass]
  public sealed class DistinctSolutionsAnalyzer : SingleSuccessorOperator, IAnalyzer {
    private const string ResultsParameterName = "Results";
    private const string PopulationGraphResultParameterName = "PopulationGraph";
    private const string SchemataParameterName = "Schemata";
    private const string DistinctSolutionsInGenerationsMatrixParameterName = "DistinctSolutionsInGenerationsMatrix";
    private const string OverallDistinctSolutionsParameterName = "OverallDistinctSolutions";
    private const string QualitiesParameterName = "Qualities";
    private const string DistanceMatrixParameterName = "DistanceMatrix";
    private const string GenerationsParameterName = "Generations";
    private const string PopulationSizeParameterName = "PopulationSize";
    private const string BestSolutionParameterName = "Best TSP Solution";
    private const string EvaluatedSolutionsParameterName = "EvaluatedSolutions";

    #region IAnalyzer Members
    public bool EnabledByDefault {
      get { return true; }
    }
    #endregion

    #region Parameter properties
    public ILookupParameter<ResultCollection> ResultsParameter {
      get { return (ILookupParameter<ResultCollection>)Parameters[ResultsParameterName]; }
    }
    public ILookupParameter<DistanceMatrix> DistanceMatrixParameter {
      get { return (ILookupParameter<DistanceMatrix>)Parameters[DistanceMatrixParameterName]; }
    }
    public ILookupParameter<IntValue> GenerationsParameter {
      get { return (ILookupParameter<IntValue>)Parameters[GenerationsParameterName]; }
    }
    public ILookupParameter<IntValue> PopulationSizeParameter {
      get { return (ILookupParameter<IntValue>)Parameters[PopulationSizeParameterName]; }
    }
    #endregion

    #region Properties
    public ResultCollection Results {
      get { return ResultsParameter.ActualValue; }
    }
    public DistanceMatrix DistanceMatrix {
      get { return DistanceMatrixParameter.ActualValue; }
    }
    public HeuristicLab.Analysis.DataTable Qualities {
      get { return ((HeuristicLab.Analysis.DataTable)ResultsParameter.ActualValue[QualitiesParameterName].Value); }
    }
    public IntValue EvaluatedSolutions {
      get { return ((IntValue)ResultsParameter.ActualValue[EvaluatedSolutionsParameterName].Value); }
    }
    public PathTSPTour BestSolution {
      get { return ((PathTSPTour)ResultsParameter.ActualValue[BestSolutionParameterName].Value); }
    }
    #endregion

    [StorableConstructor]
    private DistinctSolutionsAnalyzer(bool deserializing) : base(deserializing) { }
    private DistinctSolutionsAnalyzer(DistinctSolutionsAnalyzer original, Cloner cloner) : base(original, cloner) { }
    public DistinctSolutionsAnalyzer()
      : base() {
      Parameters.Add(new LookupParameter<ResultCollection>(ResultsParameterName, "The results collection where the analysis values should be stored."));
      Parameters.Add(new LookupParameter<DistanceMatrix>(DistanceMatrixParameterName, "The distance matrix of the TSP."));
      Parameters.Add(new LookupParameter<IntValue>(GenerationsParameterName, "Nr of generations."));
      Parameters.Add(new LookupParameter<IntValue>(PopulationSizeParameterName, "Population Size."));
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new DistinctSolutionsAnalyzer(this, cloner);
    }

    public override IOperation Apply() {
      var graph = (GenealogyGraph<Permutation>)Results[PopulationGraphResultParameterName].Value;
      int generations = GenerationsParameter.ActualValue.Value;
      int popSize = PopulationSizeParameter.ActualValue.Value;
      int evaluatedSolutions = EvaluatedSolutions.Value;
      int overallDistinctSolutions = 0;

      Dictionary<int, int> distinctSolsInGen = new Dictionary<int, int>();

      for (int i = 0; i < generations; i++) {
        var individuals = graph.Values.Where(x => x.Rank.Contains(i));
        distinctSolsInGen.Add(i, individuals.Count());
        overallDistinctSolutions += individuals.Count();
      }

      DataTable dt = new DataTable();
      DataRow row = new DataRow("Distinct Solutions");
      dt.Rows.Add(row);
      for (int i = 0; i < distinctSolsInGen.Keys.Count; i++) {
        var element = distinctSolsInGen.ElementAt(i);
        dt.Rows["Distinct Solutions"].Values.Add(element.Value);
      }

      Results.Add(new Result(DistinctSolutionsInGenerationsMatrixParameterName, dt));
      Results.Add(new Result(OverallDistinctSolutionsParameterName, new DoubleValue(evaluatedSolutions / (double)overallDistinctSolutions)));
      return base.Apply();
    }
  }
}
